#!/usr/bin/python
#
# Copyright 2009, Douglas Greiman
#
# Biomedin 217, Winter 2008-2009, Final Project
# Student name: Douglas Greiman
# SUNet id: duggelz
# Student id: 005512989
#
# This program performs a genome-wide association study on selected
# strains of rats, using phenotype data from NBRP Kyoto and genotype
# data from the European STAR consortium.
#
# Requires the following data files to already be present:
#     star_genotypes_070914/
#         alias_070914.txt
#         genotypes_seq_070914.txt
#     csv/
#         Phenotypes.csv
#
# Produces a number of intermediate datafiles with .pickle, .csv, and
# .txt file extensions.


# Module imports
import commands
import csv
import os
import pickle
import pprint
import sys
import time
import traceback
# Third-party modules
import numpy
import scipy
import scipy.stats
# Other modules in this directory
import query_rgd_biomart


class GlobalOptions:
    """Options specified on the command line"""
    preprocess = False
    process = False
    analyze1 = False
    analyze2 = False
    analyze3 = False
    filter_out_FXLE_LEXF = True
    MAX_SIGNIFICANT_SNP_P = 1e-5
    MIN_SAMPLE_SIZE = 5
    MAX_ENRICHMENT_P = 0.05
# endclass


def read_csv_table(filename, dialect, transpose):
    """Read a 2d table from a file in comma-separated value format.

    The column titles must be in the first row, and the row titles
    must be in the first column.

    Returns (row titles, column titles, table data), where each item
    is NumPy array.
    """
    infile = open(filename)
    reader = csv.reader(infile, dialect)
    lines = list(reader)
    table = numpy.array(lines)
    if transpose:
        table = table.T
    # endif
    return table[1:,0].flatten(), table[0,1:].flatten(), table[1:,1:]
# enddef


def read_csv_table_plain(filename, dialect, transpose):
    """Read a 2d table from a file in comma-separated value format.

    Unlike read_csv_table above, there are no row or column titles.
    The entire contents of the file are returned as the table.

    Returns table data as a NumPy array.
    """
    infile = open(filename)
    reader = csv.reader(infile, dialect)
    lines = list(reader)
    table = numpy.array(lines)
    if transpose:
        table = table.T
    # endif
    return table
# enddef


def pickle_arrays(filename, arrays):
    """Save a list of arrays to a file in pickle format.

    If the named file already exists, it will be overwritten.
    """
    pfile = open(filename, 'w+b')
    for array in arrays:
        array.dump(pfile)
    # endfor
    pfile.close()
# enddef


def unpickle_arrays(filename, array_count):
    """Load a tuple of arrays from a file in pickle format."""
    pfile = open(filename, 'rb')
    arrays = []
    for _ in range(array_count):
        array = numpy.load(pfile)
        arrays.append(array)
    # endfor
    return tuple(arrays)
# enddef


_mdc_id_aliases = None
def mdc_id_aliases():
    """Return the MDC lookup table.

    The table is read from a file on first use of this function.
    """
    global _mdc_id_aliases
    if _mdc_id_aliases is None:
        # Read strain name mapping table
        alias_file = open('star_genotypes_070914/alias_070914.txt')
        alias_reader = csv.DictReader(alias_file, dialect=csv.excel_tab)
        alias_lines = list(alias_reader)
        table = {}
        for line in alias_lines:
            table[line["MDC_ID"]] = line
        # endfor
        _mdc_id_aliases = table
    # endif
    return _mdc_id_aliases
# enddef


def strain_eq(nbrp_strain, mdc_id):
    """Compare an NCBP strain name and an MDC id for fuzzy equality.

    There are some minor differences in capitalization and
    nomenclature.
    
    Return True if the strains are the same, or False if not.
    """
    global _mdc_id_aliases

    nbrp_strain = nbrp_strain.lower()

    alias = mdc_id_aliases().get(mdc_id)
    if not alias:
        return False
    # endif
    star_strain = alias["Symbol"].lower()
                    
    if nbrp_strain == star_strain:
        return True
    elif ((nbrp_strain.endswith("/stm") or
           nbrp_strain.endswith("/utp")) and
          nbrp_strain[:-4] == star_strain):
        return True
    else:
        return False
    # endif
# endif


def harmonize_columns(arr1_cols, arr1, arr2_cols, arr2, col_equal,
                      new_dtype1, new_dtype2):
    """Find the column names that two arrays have in common,
    and return new arrays with only those columns present.
    """
    
    # For each column in arr1, find a single matching column in arr2
    shared_cols = []
    for arr1_col_idx in range(len(arr1_cols)):
        arr1_col = arr1_cols[arr1_col_idx]
        arr2_matches = [arr2_col_idx for arr2_col_idx in range(len(arr2_cols))
                        if col_equal(arr1_col, arr2_cols[arr2_col_idx])]
        assert len(arr2_matches) < 2, ("Ambiguous column name %s" % arr1_col)
        if len(arr2_matches) == 1:
            arr2_col_idx = arr2_matches[0]

            # Sanity check that there aren't multiple matches either way
            arr2_col = arr2_cols[arr2_col_idx]
            arr1_matches = [ci for ci in range(len(arr1_cols))
                            if col_equal(arr1_cols[ci], arr2_col)]
            assert len(arr1_matches) == 1, ("Ambiguous column name %s" %
                                            arr2_col)
            shared_cols.append((arr1_col, arr1_col_idx, arr2_col_idx))
        # endif
    # endfor

    # Create new subset arrays containing only the matching column names
    new_arr1_idxs = [shared_col[1] for shared_col in shared_cols]
    new_arr1_cols = numpy.array([arr1_cols[idx] for idx in new_arr1_idxs])
    new_arr1 = numpy.array(arr1[:,new_arr1_idxs], dtype=new_dtype1)
    new_arr2_idxs = [shared_col[2] for shared_col in shared_cols]
    new_arr2_cols = numpy.array([arr2_cols[idx] for idx in new_arr2_idxs])
    new_arr2 = numpy.array(arr2[:,new_arr2_idxs], dtype=new_dtype2)
    return (new_arr1_cols, new_arr1, new_arr2_cols, new_arr2)
# endif


def do_preprocess():
    """Merge data files from NBRP and STAR"""

    # Read the raw data from each source
    raw_phenotype_rows, raw_phenotype_cols, raw_phenotype_data = (
        read_csv_table("csv/Phenotypes.csv", csv.excel, transpose=True))
    print "Found %d phenotypes" % raw_phenotype_data.shape[0]
    pickle_arrays('raw_phenotype.pickle', (
        raw_phenotype_rows, raw_phenotype_cols, raw_phenotype_data))
    raw_genotype_rows, raw_genotype_cols, raw_genotype_data = (
        read_csv_table('star_genotypes_070914/genotypes_seq_070914.txt',
                       csv.excel_tab, transpose=False))
    print "Found %d SNPs" % raw_genotype_data.shape[0]
    pickle_arrays('raw_genotype.pickle',
                  (raw_genotype_rows, raw_genotype_cols, raw_genotype_data))


    # Fill in blanks with Inf so numpy doesn't choke
    raw_phenotype_data[raw_phenotype_data == ''] = 'inf'

    # Find matching strains between the two data sources
    (phenotype_cols,
     phenotype_data,
     genotype_cols,
     genotype_data) = harmonize_columns(raw_phenotype_cols,
                                        raw_phenotype_data,
                                        raw_genotype_cols,
                                        raw_genotype_data,
                                        strain_eq,
                                        'float', 'int')
    phenotype_rows = raw_phenotype_rows[:]
    genotype_rows = raw_genotype_rows[:]
    
    print "Found %d matching strains" % len(phenotype_cols)
    
    # Save preprocessed data for both sources
    pickle_arrays('phenotype.pickle',
                  (phenotype_rows, phenotype_cols, phenotype_data))
    pickle_arrays('genotype.pickle',
                  (genotype_rows, genotype_cols, genotype_data))
    
    # Sanity check - Validate dumped data
    (pr,pc,pd,gr,gc,gd) = load_data()
    assert numpy.all(pr == phenotype_rows)
    assert numpy.all(pc == phenotype_cols)
    assert numpy.allclose(pd, phenotype_data)
    assert numpy.all(gr == genotype_rows)
    assert numpy.all(gc == genotype_cols)
    assert numpy.allclose(gd, genotype_data)

    # debugging
    GlobalOptions.pre_process_locals = locals()
# enddef


def need_preprocess():
    """Checked whether the phenotype/genotype data has been preprocessed."""
    if os.path.exists('phenotype.pickle') and os.path.exists('genotype.pickle'):
        return False
    else:
        return True
    # endif
# enddef


def load_data():
    """Load preprocessed phenotype and genotype data."""
    (phenotype_rows, phenotype_cols, phenotype_data) = \
                     unpickle_arrays('phenotype.pickle', 3)
    (genotype_rows, genotype_cols, genotype_data) = \
                    unpickle_arrays('genotype.pickle', 3)
    return (phenotype_rows, phenotype_cols, phenotype_data,
            genotype_rows, genotype_cols, genotype_data)
# enddef


def analyze_pair(phenotype, phenotype_vec,
                 genotype, genotype_vec):
    """Analyze a single phenotype-genotype combination."""

    # Filter out missing phenotypes (these were set to +Inf in the
    # preprocessing stage.
    phenotype_mask = numpy.isfinite(phenotype_vec)
    # Filter out missing genotypes, group into the three allele classes
    genotype_masks = [(genotype_vec == 0) & phenotype_mask,
                      (genotype_vec == 1) & phenotype_mask,
                      (genotype_vec == 2) & phenotype_mask]

    # Group samples by genotype
    samples = [phenotype_vec[genotype_masks[0]],
                phenotype_vec[genotype_masks[1]],
               phenotype_vec[genotype_masks[2]]]
    sizes = [len(s) for s in samples]

    # For each of the (3 choose 2) pairs of allele classes, compare
    # the phenotype values
    test_results = []
    for i1, i2 in ((0,1),(0,2),(1,2)):
        # Only compare if the sample sizes are non-neglible
        if (sizes[i1] > GlobalOptions.MIN_SAMPLE_SIZE and
            sizes[i2] > GlobalOptions.MIN_SAMPLE_SIZE):
            try:
                # Use the Mann-Whitney rank test.  Sample sizes here
                # are small and unlikely to be normal.
                test_result = scipy.stats.mannwhitneyu(samples[i1], samples[i2])
                test_results.append(list(test_result) + [i1,i2])
            except ValueError, e:
                # Mann-Whitney raises an exception if both
                # distibutions are exactly the same, which
                # occasionally happens.
                if e.message != 'All numbers are identical in amannwhitneyu':
                    print e
                    print (phenotype, genotype, samples[i1], samples[i2])
                # endif
            # endtry
        # endif
    # endfor
    return test_results
# enddef


def do_process():
    """Analyze each possible combination of phenotype and SNP."""
    
    if need_preprocess():
        raise AssertionError("Need to preprocess data first, exiting.")
    # endif
    
    (phenotype_rows, phenotype_cols, phenotype_data,
     genotype_rows, genotype_cols, genotype_data) = load_data()

    # Ignore the highly correlated recombinant strains
    if GlobalOptions.filter_out_FXLE_LEXF:
        filter_mask = numpy.array([not (col.startswith('FXLE') or
                                        col.startswith('LEXF'))
                                   for col in phenotype_cols])
        phenotype_cols = phenotype_cols[filter_mask]
        phenotype_data = phenotype_data[:,filter_mask]
        genotype_cols = genotype_cols[filter_mask]
        genotype_data = genotype_data[:,filter_mask]
    # endif

    # Write results to these files
    results_outfile = open('results.csv', 'w+')
    results_writer = csv.writer(results_outfile)
    shuffle_outfile = open('shuffle.csv', 'w+')
    shuffle_writer = csv.writer(shuffle_outfile)

    # Iterate over all pairs of (Phenotype, SNP)
    all_test_results = []
    all_shuffle_results = []
    total_test_count = len(phenotype_rows) * len(genotype_rows)
    start_time = time.time()
    test_num = 0
    for phenotype_idx in range(len(phenotype_rows)):
        phenotype = phenotype_rows[phenotype_idx]
        phenotype_vec = phenotype_data[phenotype_idx,:]
        for genotype_idx in range(len(genotype_rows)):
            genotype = genotype_rows[genotype_idx]
            genotype_vec = genotype_data[genotype_idx,:]

            assert len(phenotype_vec) == len(genotype_vec)

            # Process snp/phenotype pair
            test_results = analyze_pair(phenotype, phenotype_vec,
                                        genotype, genotype_vec)
            for result in test_results:
                row = (result[1], phenotype, genotype) + tuple(result)
                all_test_results.append(row)
                results_writer.writerow(row)
            # endfor

            # Permute phenotypes and process again
            shuffled_vec = phenotype_vec.copy()
            numpy.random.shuffle(shuffled_vec)
            shuffle_results = analyze_pair(
                phenotype, shuffled_vec,
                genotype, genotype_vec)
            for result in shuffle_results:
                row = (result[1], phenotype, genotype) + tuple(result)
                all_shuffle_results.append(row)
                shuffle_writer.writerow(row)
            # endfor

            # Progress report to the user
            if test_num % 1000 == 1:
                print "Starting test %d of %d" % (test_num, total_test_count)
                elapsed_time = time.time() - start_time
                rate = test_num / elapsed_time
                print "Elapsed time %.0f seconds, %.0f seconds remaining (%.0f tests per sec)" % (
                    elapsed_time, (total_test_count - test_num) / rate, rate)
            # endif
            test_num += 1
        # endfor
    # endfor

    # Cleanup
    results_outfile.close()
    shuffle_outfile.close()

    # debugging
    GlobalOptions.process_locals = locals()
# enddef


def do_analyze1():
    """Find significant phenotype/SNP associations."""

    # Read the results and shuffled-phenotype results
    atr = read_csv_table_plain("results.csv", csv.excel, transpose=False)
    asr = read_csv_table_plain("shuffle.csv", csv.excel, transpose=False)

    # Filter out results above p-value cutoff
    p_cutoff = GlobalOptions.MAX_SIGNIFICANT_SNP_P
    atr_p = numpy.array(atr[:,0], dtype="float")
    asr_p = numpy.array(asr[:,0], dtype="float")
    atr5 = atr[atr_p < p_cutoff,:]
    asr5 = asr[asr_p < p_cutoff,:]
    print "At p-value cutoff %f, had %d results and %d shuffle results" % (
        p_cutoff, len(atr5), len(asr5))

    pickle_arrays("sig_results.pickle", (atr5, asr5))

    # debugging
    GlobalOptions.analyze1_locals = locals()
# enddef


def count_distinct(seq, min=None):
    """Count the number of distinct items in a sequence.

    seq: A sequence
    min: If not None, only return items that occur min or more times
    
    Returns a map of {item: count of occurrences}.
    """
    d = {}
    for item in seq:
        d.setdefault(item, 0)
        d[item] += 1
    # endfor

    # Filter out infrequent items
    if min is not None:
        d = dict([(k,v) for k,v in d.items() if v >= min])
    # endif
    
    return d
# enddef


def intx(i):
    """Convert a string to integer if possible.

    i: A string
    
    Returns an integer if possible, otherwise returns original string.
    """
    
    try:
        return int(i)
    except ValueError:
        return i
    # endtry
# enddef


def gather(key_vec, value_vec):
    """Gather value_vec into lists based on keys in key_vec."""
    assert len(key_vec) == len(value_vec)
    d = {}
    for idx in range(len(key_vec)):
        d.setdefault(key_vec[idx],[]).append(value_vec[idx])
    # endfor
    return d
# enddef


def do_analyze2():
    """Get Gene Ontology terms for every SNP tested from the RGD website"""
    
    # Get meta-data for snps
    (raw_genotype_rows, raw_genotype_cols, raw_genotype_data) = (
        unpickle_arrays('raw_genotype.pickle', 3))
    raw_snp_map = {}
    for row_idx in range(len(raw_genotype_rows)):
        raw_snp_map[raw_genotype_rows[row_idx]] = raw_genotype_data[row_idx,:]
    # endfor

    # Query the RGD
    start_time = time.time()
    snp_num = 0
    snp_rgd_data = {}
    for snp in raw_genotype_rows:
        snp_row = raw_snp_map[snp]
        chromosome = intx(snp_row[0])
        location = int(snp_row[1])
        try:
            rgd_data = query_rgd_biomart.query(chromosome, location)
            snp_rgd_data[snp] = (snp, chromosome, location, rgd_data)
        except StandardError:
            traceback.print_exc()
        # endtry

        # Progress report to the user
        if snp_num % 100 == 1:
            print "Starting snp %d of %d" % (snp_num, len(raw_genotype_rows))
            elapsed_time = time.time() - start_time
            rate = snp_num / elapsed_time
            print "Elapsed time %.0f seconds, %.0f seconds remaining (%.1f secs per snp)" % (
                elapsed_time, (len(raw_genotype_rows) - snp_num)/rate, 1/rate)

            # Save partial progress
            pfile = open('snp_rgd_map.pickle', 'w+b')
            pickle.dump(snp_rgd_data, pfile)
            pfile.close()
        # endif
        snp_num += 1
    # endfor

    pfile = open('snp_rgd_map.pickle', 'w+b')
    pickle.dump(snp_rgd_data, pfile)
    pfile.close()

    # debugging
    GlobalOptions.analyze2_locals = locals()
# enddef


def do_analyze3():
    """Final analysis"""
    atr5, asr5 = unpickle_arrays("sig_results.pickle", 2)
    atr5_phen = atr5[:,1]
    atr5_snp = atr5[:,2]

    p_and_snps_by_phenotype_map = {}
    for rowidx in range(len(atr5)):
        p, phen, snp = atr5[rowidx,[0,1,2]]
        p = float(p)
        p_and_snps_by_phenotype_map.setdefault(phen,[])
        p_and_snps_by_phenotype_map[phen].append((p, snp))
    # endfor
    for pairs in p_and_snps_by_phenotype_map.values():
        pairs.sort()
    # endfor
    #pprint.pprint(p_and_snps_by_phenotype_map)
    
    # Group and count by SNP
    snp_counts = count_distinct(atr5_snp)
    print "Significant SNPs by counts:"
    pprint.pprint(snp_counts)
    print

    # Load RGD data
    pfile = open('snp_rgd_map.pickle', 'rb')
    snp_rgd_map = pickle.load(pfile)
    pfile.close()

    # Show bp positions of significant SNPs
    snp_pos = [(intx(snp_rgd_map[snp][1]), int(snp_rgd_map[snp][2]))
               for snp in snp_counts.keys()]
    print "Significant SNPs by position:"
    pprint.pprint(sorted(snp_pos))
    print

    # Group by phenotype
    snps_by_phenotype_map = gather(atr5_phen, atr5_snp)

    # go_ref := go_id + go_term

    # Get counts for overall total number of go terms, and overall
    # count for each go term
    overall_total_go_ref_count = 0
    overall_go_ref_counts = {}
    overall_symbol_list = {}
    overall_entrez_gene_id_list = {}
    overall_rgd_id_list = {}
    for snp, value in snp_rgd_map.items():
        for row in value[3]:
            go_id = row["go_id"]
            go_term = row["go_term"]
            go_ref = "%s(%s)" % (go_id, go_term)
            overall_go_ref_counts.setdefault(go_ref, 0)
            overall_go_ref_counts[go_ref] += 1
            overall_total_go_ref_count += 1
            symbol = row["symbol"]
            if symbol:
                overall_symbol_list[symbol] = 1
            # endif
            entrez_gene_id = row["entrez_gene_id"]
            if entrez_gene_id:
                overall_entrez_gene_id_list[entrez_gene_id] = 1
            # endif
            rgd_id = row["rgd_id"]
            if rgd_id:
                overall_rgd_id_list[rgd_id] = 1
            # endif
        # endfor
    # endfor

    # Save overall counts
    f = open('symbols.txt', 'w+')
    f.write(' '.join(overall_symbol_list.keys()))
    f.close()
    f = open('entrez_gene_ids.txt', 'w+')
    f.write(' '.join(overall_entrez_gene_id_list.keys()))
    f.close()
    f = open('rgd_ids.txt', 'w+')
    f.write(' '.join(overall_rgd_id_list.keys()))
    f.close()
    
    # GO terms by phenotype
    sorted_phenotype_list = snps_by_phenotype_map.keys()
    sorted_phenotype_list.sort()
    for phenotype in sorted_phenotype_list:
        snps = snps_by_phenotype_map[phenotype]
        p_and_snps = p_and_snps_by_phenotype_map[phenotype]
        #print "Phenotype: %s, SNPS: %s" % (phenotype, snps)

        # Count # of mentions of each go term for this phenotype
        this_phenotype_total_go_ref_count = 0
        this_phenotype_go_ref_counts = {}
        this_phenotype_symbol_list = {}
        this_phenotype_entrez_gene_id_list = {}
        this_phenotype_rgd_id_list = {}
        for snp in snps:
            for row in snp_rgd_map[snp][3]:
                go_id = row["go_id"]
                go_term = row["go_term"]
                go_ref = "%s(%s)" % (go_id, go_term)
                if go_id:
                    this_phenotype_go_ref_counts.setdefault(go_ref, 0)
                    this_phenotype_go_ref_counts[go_ref] += 1
                    this_phenotype_total_go_ref_count += 1
                # endif

                symbol = row["symbol"]
                if symbol:
                    this_phenotype_symbol_list[symbol] = 1
                # endif
                entrez_gene_id = row["entrez_gene_id"]
                if entrez_gene_id:
                    this_phenotype_entrez_gene_id_list[entrez_gene_id] = 1
                # endif
                rgd_id = row["rgd_id"]
                if rgd_id:
                    this_phenotype_rgd_id_list[rgd_id] = 1
                # endif
            # endfor
        # endfor

        # Calculate hypergeometric p for each go term mentioned for
        # this phenotype
        go_ref_ps = []
        for go_ref, this_phenotype_go_ref_count in this_phenotype_go_ref_counts.items():
            overall_go_ref_count = overall_go_ref_counts[go_ref]
            p = go_term_enrichment_p(
                this_phenotype_go_ref_count, overall_total_go_ref_count,
                overall_go_ref_count, this_phenotype_total_go_ref_count)
            if p < GlobalOptions.MAX_ENRICHMENT_P:
                go_ref_ps.append([p, go_ref, this_phenotype_go_ref_count])
            # endif
        # endfor

        # Print results for this phenotype
        print '\nPhenotype: "%s"' % phenotype

        print '  SNPs:'
        p_and_snps = p_and_snps_by_phenotype_map[phenotype]
        p_and_snps.sort()
        for p, snp in p_and_snps:
            print '\t%.8f  %s' % (p, snp)
        # endfor
        
        print '  Symbols: %s' % ' '.join(this_phenotype_symbol_list.keys())
        print '  Entrez Gene Ids: %s' % ' '.join(this_phenotype_entrez_gene_id_list.keys())
        print '  RGD Ids: %s' % ' '.join(this_phenotype_rgd_id_list.keys())

        print '  Enriched GO Terms:'
        go_ref_ps.sort(lambda a,b: cmp(a[0],b[0]) or cmp(a[1:],b[1:]))
        for p, go_ref, go_ref_count in go_ref_ps:
            print "\t%.8f [%d of %d] %s" % (
                p, go_ref_count,
                overall_go_ref_counts[go_ref], go_ref)
        # endfor

        print
        for p, snp in p_and_snps[0:3]:
            print snp
        # endfor
        print
        for p, go_ref, go_ref_count in go_ref_ps[0:3]:
            print go_ref
        # endfor
        
    # endfor

    # debugging
    GlobalOptions.analyze3_locals = locals()
# enddef

# Models drawing objects from a bin. 
#def hypergeom_cdf(k, M, n, N):
def go_term_enrichment_p(k, M, n, N):
    """Compute go term enrichment probability for this term.

    M is total number of objects, n is total number of Type I
    objects. k is number of Type I objects in N drawn without
    replacement from population.
    """
    p = 0
    for i in range(k-1, -1, -1):
        # Apparently this is buggy in scipy
        #pmf = scipy.stats.hypergeom.pmf(i, M, n, N)

        # So use R instead (slow, since we spawn a new process for
        # each call)
        status, output = commands.getstatusoutput(
            "echo 'dhyper(%d,%d,%d,%d)'|R --slave" % (
            i, n, M-n, N))
        assert status == 0, (status, output)
        words = output.split()
        assert len(words) == 2, output
        pmf_r = float(words[1])
        #print "PMFs: %f %f" % (pmf, pmf_r)
        
        p += pmf_r
    # endfor

    # Clamp to non-negative
    return max(0, 1-p)
# enddef


def usage():
    """Print command line arguments for this program."""
    print >> sys.stderr, \
          "Usage: %s [--preprocess] [--process] [--analyze1|2|3]" % sys.argv[0]
# enddef


def main():
    """Main entry point for this program."""

    # Check command line arguments
    if len(sys.argv) == 1:
        usage()
        sys.exit(1)
    # endif
    for arg in sys.argv[1:]:
        # Allow any item in GlobalOptions to be specified on the command line
        if (arg.startswith('--') and
            not arg.startswith('--_') and
            arg[2:] in dir(GlobalOptions)):
            setattr(GlobalOptions, arg[2:], True)
        else:
            usage()
            sys.exit(1)
        # endif
    # endfor

    # Excute specified steps of the analysis, in order
    if GlobalOptions.preprocess:
        do_preprocess()
    #endif
    if GlobalOptions.process:
        do_process()
    # endif
    if GlobalOptions.analyze1:
        do_analyze1()
    # endif
    if GlobalOptions.analyze2:
        do_analyze2()
    # endif
    if GlobalOptions.analyze3:
        do_analyze3()
    # endif
#enddef

if __name__ == '__main__':
    main()
