#!/usr/bin/python
#
# Copyright 2009, Douglas Greiman
#
# Biomedin 217, Winter 2008-2009, Final Project
# Student name: Douglas Greiman
# SUNet id: duggelz
# Student id: 005512989
#
# This program parses web pages downloaded from the NBRP Rat (Kyoto)
# web site, and extracts all the relevant data into in a single data
# file in CSV format.
#
# Requires the following data files to already be present:
#     data_NBRP_Rat_Kyoto/
#         act_list.aspx_{1,2,3}.html
#         biochemblood_list.aspx_{1,2,3}.html
#         BP_BT_list.aspx_{1,2,3}.html
#         bw_list.aspx_{1,2,3}.html
#         hematology_list.aspx_{1,2,3}.html
#         OW_list.aspx_{1,2,3}.html
#         pat_list.aspx_{1,2,3}.html
#         urine_list.aspx_{1,2,3}.html
#
# Produces a number of files in the csv/ directory with .csv file
# extensions.  The final output is in csv/Phenotypes.csv

# Module imports
import csv
import os
import pprint
import re
# Third-party imports
from BeautifulSoup import BeautifulSoup  # HTML parser ("tag soup")

# Classes
class Descriptor:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)
    # enddef
    def __str__(self):
        return pprint.pformat(self.__dict__)
    # enddef
# endclass

class PageSet:
    """A single 2-d table of data spread over several web pages."""
    def __init__(self, descriptor):
        self.descriptor = descriptor
        self.header_row = None  # List of cells
        self.data = None  # List of rows, each row is list of cells
    # enddef
# endclass


# Constants
DATA_DIR = "data_NBRP_Rat_Kyoto"
PAGESET_DESCRIPTORS = [
    Descriptor(name="Blood Pressure and Body Temperature",
               template="BP_BT_list.aspx_%d.html", last_page=3),
    Descriptor(name="Blood Biochemistry",
               template="biochemblood_list.aspx_%d.html", last_page=3),
    Descriptor(name="Hematology",
               template="hematology_list.aspx_%d.html", last_page=3),
    Descriptor(name="Urine Parameters",
               template="urine_list.aspx_%d.html", last_page=3),
    Descriptor(name="Body Weights",
               template="bw_list.aspx_%d.html", last_page=3),
    Descriptor(name="Organ Weights",
               template="OW_list.aspx_%d.html", last_page=3),
    Descriptor(name="Locomotor Activity",
               template="act_list.aspx_%d.html", last_page=3),
    Descriptor(name="Passive Avoidance Test",
               template="pat_list.aspx_%d.html", last_page=3),
    ]


# Functions
def scrape_all(pageset_descriptors):
    """Scrape all NBRP webpages and construct master data tables."""
    pagesets = []
    for descriptor in pageset_descriptors:
        pageset = PageSet(descriptor)
        print "Scanning pageset %s" % pageset.descriptor.name
        pageset.header_row, pageset.data = scrape_pageset(
            pageset.descriptor.template, pageset.descriptor.last_page)
        #pprint.pprint(pageset.header_row)
        #pprint.pprint(pageset.data)
        pagesets.append(pageset)
    # endfor
    return pagesets
# enddef


def scrape_pageset(page_template, last_page):
    """Scrape pages 1 to 'last_page' of a multi-page NBRP table."""
    pageset_data = []
    header_row = None
    for page_number in range(1, last_page+1):
        filename = os.path.join(DATA_DIR, page_template % page_number)
        page_header_row, page_data = scrape_page(filename)
        if header_row is None:
            header_row = page_header_row
        else:
            assert page_header_row == header_row, (page_template, page_number)
        # endif
        pageset_data.extend(page_data)
    # endfor
    return header_row, pageset_data
# enddef


def scrape_page(page_filename):
    """Extract a table of data from an NBRP webpage."""
    # Open and parse filename
    infile = open(page_filename)
    soup = BeautifulSoup(infile)

    # Data is in 3rd table, or 2nd with css class
    tables = soup.findAll('table', attrs={'class': "MultipadsFormTABLE"})
    assert len(tables) == 2
    table = tables[1]
    trs = table.findAll('tr')
    assert len(trs) > 0
    page_data = []

    # First row is column headers.
    header_tr = trs[0]
    header_row = []
    for td in header_tr.findAll('td'):
        text = td.findAll(text=True)
        header = html_text_to_data(text)
        header_row.append(header)
    # endfor
  
    # Last row is "goto page X" navigation
    assert len(trs[-1].findAll('td')) == 1

    # Iterate over all data rows
    for tr in trs[1:-1]:
        row = []
        for td in tr.findAll('td'):
            text = td.findAll(text=True)
            data = html_text_to_data(text)
            row.append(data)
        # endfor
        page_data.append(row)
    # endfor

    return (header_row, page_data)
# enddef


def html_text_to_data(text):
    """Convert a string of html cdata into a Python data object.

    Also converts Unicode strings to UTF-8 encoded regular strings.
    """
    data = ' '.join([item.replace('&nbsp;','').replace('&nbsp','').strip()
                    for item in text])
    data = re.sub('\\s+', ' ', data).strip()
    data = data.encode('utf-8')
    try:
        data = int(data)
    except ValueError:
        try:
            data = float(data)
        except:
            pass
        # endtry
    # endtry
    return data
# enddef


def dump_pageset_as_csv(pageset):
    """Save the contents of the pageset as a single CSV data file"""
    outfilename = pageset.descriptor.name.replace(' ', '_') + '.csv'
    outfile = open(os.path.join("csv", outfilename), 'wb')
    writer = csv.writer(outfile)
    writer.writerow(pageset.header_row)
    for row in pageset.data:
        writer.writerow(row)
    # endfor
# enddef


def main():
    # Scrape data
    pagesets = scrape_all(PAGESET_DESCRIPTORS)
    
    # Do some validation and sanitizing
    for pageset in pagesets:
        pageset.data.sort()
        for row_idx in range(len(pageset.data)):
            row = pageset.data[row_idx]
            # Fixup a mistake in the strain name
            if ((row[0] == 'FXLE23/Stm') and
                (row_idx > 1) and
                (pageset.data[row_idx-1][0] == row[0])):
                row[0] = 'FXLE23/Stm(F)'
            # endif
            for cell_idx in range(0, len(row)):
                cell = row[cell_idx]
                if cell_idx > 0:
                    assert (cell == "" or
                            isinstance(cell, int) or
                            isinstance(cell, float)), (row_idx, cell_idx)
                # endif
            # endfor
        # endfor

        # Dump to file
        dump_pageset_as_csv(pageset)
    # endfor

    # Create master list of all column names
    master_col_names = []
    master_col_idxs = {}
    for pageset in pagesets:
        for col_name in pageset.header_row:
            if col_name not in master_col_names:
                master_col_idxs[col_name] = len(master_col_names)
                master_col_names.append(col_name)
            # endif
        # endfor
    # endfor

    # Create master list of all row names
    master_row_names = []
    master_row_idxs = {}
    for pageset in pagesets:
        for row in pageset.data:
            row_name = row[0]
            if row_name not in master_row_names:
                master_row_idxs[row_name] = len(master_row_names)
                master_row_names.append(row_name)
            # endif
        # endfor
    # endfor

    # Merge all pagesets into a single table
    master = PageSet(Descriptor(name="Phenotypes"))
    master.header_row = master_col_names[:]
    master.data = []
    for row_idx in range(len(master_row_names)):
        row = [None] * len(master_col_names)
        master.data.append(row)
    # endfor
    for pageset in pagesets:
        for row_idx in range(len(pageset.data)):
            row = pageset.data[row_idx]
            row_name = row[0]
            master_row_idx = master_row_idxs[row_name]

            for col_idx in range(len(row)):
                col_name = pageset.header_row[col_idx]
                master_col_idx = master_col_idxs[col_name]
                data = row[col_idx]

                existing_data = master.data[master_row_idx][master_col_idx]
                if existing_data is not None and data != existing_data:
                    print ("Consistency error: Two different measurements for "
                           "%s:%s=[%s, %s] in pageset %s (%s,%s,%s,%s)") % (
                        row_name, col_name,
                        data, existing_data,
                        pageset.descriptor.name, row_idx, col_idx,
                        master_row_idx, master_col_idx)
                else:
                    master.data[master_row_idx][master_col_idx] = data
                # endif
            # endfor
        # endfor
    # #endfor

    master.data.sort()

    dump_pageset_as_csv(master)
     
    return pagesets
# enddef

if __name__ == '__main__':
    pagesets = main()
